(*
#****m* SecDec/loop/src/deco/calcFU.m
# NAME
#  calcFU.m
# USAGE
#  called from subdir/graph/graph.m, via makeFU.pl
# INPUT: 
#        momlist:  list of loop momenta
#         proplist:   list of propagators
#	  numerator:  list of  contracted loop momenta in numerator                             
#	  repl:        list of replacements                                  
#
# PURPOSE
#  calculates the functions  F, U and N (numerator)
# makes some default replacements for the invariants, and in addition uses the 
# onshell conditions defined in Template.m
# The functions F,U,N are written to subdir/graph/FUN.m
#****
*)

(* conventions p_i^2=s_i=ssp[i], (p_i+p_j)^2=s_{ij}=sp[i,j], m_i^2=ms[i],
  all momenta incoming *)


(* default replacements for general diagrams if repl is not given by the user: *)

offshell[Nmax_]:=Table[ToExpression[StringJoin["p",ToString[i]]]^2->ssp[i],{i,Nmax}];

mand[Nmax_]:=Flatten[Table[Table[ToExpression[StringJoin["p",ToString[j]]]*ToExpression[StringJoin["p",ToString[i]]]->
(sp[j,i]-ssp[j]-ssp[i])/2,{j,i-1}],{i,Nmax}]];

replgen[Nmax_]:=Join[offshell[Nmax],mand[Nmax]];



(* ************** Module which will be called from main program: ******************** *)
(* returns  functions U,F,Num,rank,ltilde  as a list , ltilde is needed for tensor integrals only        *)
(* ltilde is a vector, see eq.(8) of Int.J.Mod.Phys.A23:1457 (2008)   *)

(* NOTE: for non-planar boxes, replacement s13->-s12-s23 can NOT be done 
                because it assumes physical kinematics  
		However, naive calculation of F[z] yields a result which differs from 
		the one obtained by "cutting rules" by a term ~ (s12+s23+s13).
		This additional term leads to the wrong result (because some poles are not seen)
		for unphysical kinematics.
		The solution is to first make the replacement s13->-s12-s23
		and then undo it for all invariants coming with a negative sign.
		This leads to the result obtained by "cutting rules"
		*)

calcFU[momlist_,proplist_,numerator_,repl_,externalegs_]:=Module[{loops,props,deno,k,
                        reploc,rem,ex1,ex2,ex3,exn1,exn2,co1,coij,tmu1,tmu2,coim,con1,conij,rankall,
			he,Mm,Nvec,Fhe,Ff,Uu,Num,ltilde,Fh1,Fh2,Fh3,Fh4,Fh5,
			listcoefki,external,r,ltil,tensorprefac,dotloc,lvec,le,
			mulist,ext,contractextint,prefaclist,prenumb,
			nontensornumerator,loopmomcheck,kchanged,indexlist,inctens2,inctens,tenslist,hgchanged},


reploc=Join[repl,replgen[externalegs]];

loops=Length[momlist];
props=Length[proplist];
deno=Sum[z[i]*proplist[[i]],{i,props}];


(* needed for numerator: *)

(* initialize *)
rankall=0;


(* construct list of contracting (external) vectors for each loop momentum *)

Do[listcoefki[i]={},{i,loops}];
external={};
Do[r[i]=0,{i,Nn}];
nontensornumerator=1;
loopmomcheck=Table[momlist[[i]]->0,{i,Length[momlist]}];

Do[
 kchanged=False;
  For[i=1,i<=loops,i++,
     exn1=Expand[numerator[[k]], momlist[[i]]];
      For[j=1,j<=i,j++,
     exn2=Expand[exn1,momlist[[j]]];
      conij[i,j]=Coefficient[exn2, momlist[[i]]*momlist[[j]]];
      If[conij[i,j]=!=0 , 
        kchanged=True;
	r[i]=r[i]+1;
        tmu1=mu[i,r[i]];
	r[j]=r[j]+1;
	tmu2=mu[j,r[j]];
        external=Append[external,hgmn[tmu1,tmu2]];
        ];
     ]; (* end loop over j *)
      con1=Coefficient[exn1,momlist[[i]]];
      If[MatchQ[con1/.loopmomcheck,0]==False,
	kchanged=True;
	r[i]=r[i]+1;
        listcoefki[i]=Append[listcoefki[i],con1];
	external=Append[external,extvec[con1,mu[i,r[i]]]];
       ];
      ]; (* end loop over i *)
      If[kchanged==False,nontensornumerator=nontensornumerator*(numerator[[k]]//.reploc)];
   ,{k,Length[numerator]}];

(* the mu[i,r[i]]] are the Gamma_i: r is r-th Lorentz index belonging to the ith loop momentum *)

Do[looprank[i]=r[i],{i,loops}];
rankall=Sum[looprank[i],{i,loops}];
pp=Floor[rankall/2];


(* *********** Denominator ************* *)

rem=deno;
For[i=1,i<=loops,i++,
 ex1=Expand[rem, momlist[[i]]];
 For[j=1,j<=loops,j++,
   ex2=Expand[ex1,momlist[[j]]];
   coij[i,j]=Coefficient[ex2, momlist[[i]]*momlist[[j]]];
   If[i<j , coim[i,j]=1/2*coij[i,j],
              coim[i,j]=coij[i,j] ];
   rem=Factor[rem-coim[i,j]*momlist[[i]]*momlist[[j]]];
 ];
]; 
Mm=Table[Table[coim[i,j],{i,loops}],{j,loops}];
For[i=1,i<=loops,i++,
 ex3=Expand[rem, momlist[[i]]];
 co1[i]=Coefficient[ex3, momlist[[i]],1];
 rem=rem-co1[i]*momlist[[i]];
];
rem=Expand[rem];
Nvec=Cancel[-1/2*Table[co1[i],{i,loops}]];
Uu=Collect[Det[Mm],Table[z[i],{i,props}]];
Fhe=Cancel[Uu*Inverse[Mm]];
ltilde=Simplify[Expand[Fhe.Nvec]];
Fh1=Expand[Nvec.Fhe.Nvec-rem*Uu];

If[externalegs==4,
  Fh2=Expand[-Fh1//.reploc/.{sp[1,4]->sp[2,3], sp[2,4]->sp[1,3],sp[3,4]->sp[1,2]}];
  Fh3=Expand[Fh2/.{sp[1,3]->-sp[2,3]-sp[1,2]+Sum[ssp[i],{i,4}]}];
  Fh4=Fh3//.{Times[-1, sp[1, 2],a__]->Times[sp[2, 3],a]+Times[sp[1,3],a]-Times[Sum[ssp[i],{i,4}],a]};
  Ff=Simplify[-Fh4//.reploc];
  ,(* else *)
  If[externalegs==5,
  Fh2=Expand[Simplify[-Fh1//.reploc]]//.{Times[-1, sp[1, 2],a__]->Times[sp[1,5],a] + Times[sp[2,5],a] - Times[sp[3,4],a]};
  Fh3=Expand[Fh2/.{Times[-1, sp[2, 3],a__]->Times[sp[1,2],a] + Times[sp[1,3],a] - Times[sp[4,5],a]}];
  Fh4=Expand[Fh3/.{Times[-1, sp[1, 5],a__]->Times[sp[1,4],a] - Times[sp[2,3],a] + Times[sp[4,5],a]}];
  Fh5=Expand[Fh4/.{Times[-1, sp[3,4],a__]->Times[sp[3,5],a] + Times[sp[4,5],a] - Times[sp[1,2],a]}];
  Ff=Simplify[-Fh5],(* else *)
  Ff=Simplify[Fh1//.reploc];
  ]; (* end if externalegs=5 *)

]; (* end if externalegs=4 *)

(* ************* Numerator: ********************* *)

ltil[kloop_]:=ltilde[[kloop]];
tensorprefac[dd_]:=Module[{res},
pp=Simplify[(rankall-dd)]/2;
res=(-1/2)^pp/(Product[Simplify[sumpow-loops*Dim/2-m],{m,1,pp}])*(Ff//.reploc)^pp;
Return[res];
];

dotloc[x_,y__]:=Module[{ex},
ex=Expand[x*y]//.reploc;
Return[ex]
];

If[rankall>0, (*tensor structure in numerator*)

mulist=Flatten[Table[Table[mu[i,r],{r,looprank[i]}],{i,loops}]];
indexlist=allcombs[mulist];
inctens2[{ic2_}]:=lt[ic2];
inctens2[{ic2a_,ic2b_}]:=hgmn[ic2a,ic2b];
inctens[inclist_]:=Times@@(inctens2/@inclist);
tenslist=inctens/@indexlist;
tenslist=tenslist//.hgmn[mu[a_,b_],mu[c_,d_]]->Fhe[[a,c]]hgchanged[mu[a,b],mu[c,d]];
tenslist=tenslist//.hgchanged->hgmn;
ext=Times@@external;
contractextint=contractmunu[ext,tenslist];
prefaclist=whichprefactor/@contractextint;
prenumb=(contractextint.prefaclist)/.{extdot->dot, intdot->dot};
prenumb=prenumb//.{lt->ltil}/.{dot -> dotloc};
Num=nontensornumerator*Factor[prenumb/.{tensorprefactor->tensorprefac}];
,
Num=nontensornumerator
];  (* endif rankall > 0 *)
he={Uu,Ff,Num,rankall,ltilde}/.{0.->0};
Return[he]
];



calcFUcut[graphlistin_,numexternallines_,numloops_,onshell_:none]:=Module[{numvert,extprops,
     graphlist,nps,vertjoin,vertmatrix,internalprops,posstreecuts,poss2treecuts,
					validtreecut,valid2treecut,treecuts,twotreecuts,tt,NOTATREE,NOTA2TREE,U,F,F0,Fmass,Fterm},
sp[a_]:=ssp[a];
If[numexternallines==4,sp[1,4]=sp[2,3]];
numvert=Max[#[[2]]&/@graphlistin];
extprops=Table[{nul,{i,i+numvert}},{i,numexternallines}];
numvert=numvert+numexternallines;
graphlist=Join[extprops,graphlistin];


nps=Length[graphlist];
Attributes[vertjoin]={Orderless};
Do[
 If[i==j,vertjoin[i,j]=1,
 vertjoin[i,j]=Sum[If[Or@@(graphlist[[k,2]]==#&/@{{i,j},{j,i}}),pr[k],0],{k,nps}]]
	,{i,numvert},{j,i,numvert}];
vertmatrix=Table[vertjoin[i,j],{i,numvert},{j,numvert}];
(* vertmatrix is an un-normalised transition matrix representation of the graph*)


internalprops=Table[i,{i,numexternallines+1,nps}];
posstreecuts=Subsets[internalprops,{numloops}]; (* all subsets of the internal propagators*)
poss2treecuts=Subsets[internalprops,{numloops+1}];(* with the required length to define a tree/2-tree resp*)

validtreecut[vlist_List]:=Module[{cutrep,newmatrix,test},
 (* checks whether a given set of propagators, when removed from the graph, define a tree*)
	cutrep=pr[#]->0&/@vlist;
	newmatrix=(vertmatrix/.cutrep)/.pr[_]->1;
	(* newmatrix is the transition matrix of the graph with the specified propagators removed *)
	(* the specified propagators define a tree iff the resulting graph is connected *)
	Do[newmatrix=newmatrix.newmatrix,{i,Floor[Log[2,numvert]+1]}];
	(* if newmatrix^numvert has any zero entries, then the graph is unconnected, and thus not a tree*)
	test=And@@(MatchQ[#,0]==False&/@Flatten[newmatrix]);
	If[test,Return[vlist],Return[NOTATREE]]];

valid2treecut[vlist_List]:=Module[{cutrep,newmatrix,j21,nj21,nj21i,valid2tree},
 (* checks whether a given set of propagators, when removed from the graph, define a 2-tree*)
	cutrep=pr[#]->0&/@vlist;
	newmatrix=(vertmatrix/.cutrep)/.pr[_]->1;
		(* newmatrix is the transition matrix of the graph with the specified propagators removed *)
	(* the specified propagators define a 2-tree iff the resulting graph has exactly 2 connected components *)
	Do[newmatrix=newmatrix.newmatrix,{i,Floor[Log[2,numvert]+1]}];
 j21={1};nj21={};nj21i={};
	Do[If[MatchQ[newmatrix[[1,j]],0],nj21=Append[nj21,j],j21=Append[j21,j]],{j,2,numexternallines}];
	Do[If[MatchQ[newmatrix[[1,j]],0],nj21i=Append[nj21i,j]],{j,numexternallines+1,numvert}];
	valid2tree=True;
	(* the propagators define a 2-tree iff all vertices not joined to vertex 1 are joined to each other *)
	If[nj21=={},valid2tree=False,
	 testvertex=nj21[[1]];valid2tree=And@@(MatchQ[newmatrix[[testvertex,#]],0]==False&/@Join[nj21,nj21i])];
	If[valid2tree,Return[{vlist,If[Length[j21]<=Length[nj21],j21,nj21]}],Return[NOTA2TREE]]];
 

treecuts=(validtreecut/@posstreecuts)//.{AA___,NOTATREE,BB___}->{AA,BB};
twotreecuts=(valid2treecut/@poss2treecuts)//.{AA___,NOTA2TREE,BB___}->{AA,BB};
Attributes[tt]={Listable};
U=Plus@@(Times@@tt[#]&/@treecuts);
Fterm[flist_List]:=-sp@@flist[[2]]*Times@@(tt[flist[[1]]]);
F0=Plus@@(Fterm/@twotreecuts);
Fmass=U Sum[graphlist[[i,1]]tt[i],{i,numexternallines+1,nps}];
F=F0+Fmass;
tt[a_Integer]:=z[a-numexternallines];
If[MatchQ[onshell,none]==False,F=F/.onshell];
Return[{U,F,1,0}]];









